/*
 * This file is part of Cockpit.
 *
 * Copyright (C) 2019 Red Hat, Inc.
 *
 * Cockpit is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * Cockpit is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Cockpit; If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include "client-certificate.h"

#include <assert.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <regex.h>
#include <stdbool.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <syslog.h>
#include <unistd.h>

#include <systemd/sd-bus.h>

#include "session-utils.h"

#define CLIENT_CERTIFICATE_DIRECTORY   "/run/cockpit/tls/clients"

static int
open_proc_pid (pid_t pid)
{
    char path[100];
    int r = snprintf (path, sizeof (path), "/proc/%d", pid);
    if (r < 0 || r >= sizeof (path))
        errx (EX, "memory error");
    int fd = open (path, O_DIRECTORY | O_NOFOLLOW | O_PATH);
    if (fd < 0)
      err (EX,  "failed to open %s", path);
    return fd;
}

/* This is a bit lame, but having a hard limit on peer certificates is
 * desirable: Let's not get DoSed by huge certs */
#define MAX_PEER_CERT_SIZE 100000

/* Reads the cgroupsv2-style /proc/[pid]/cgroup file of the process,
 * including "0::" prefix and newline.
 * NB: the kernel doesn't allow newlines in cgroup names. */
static char *
read_proc_pid_cgroup (int dirfd, size_t *out_length)
{
  int fd = openat (dirfd, "cgroup", O_RDONLY | O_CLOEXEC);
  if (fd < 0)
    {
      warn ("Failed to open /proc/pid/cgroup");
      return NULL;
    }

  char buffer[1024];
  int ret;
  while (true)
    {
      ret = read (fd, buffer, sizeof (buffer));
      if (ret < 0)
        {
          if (errno == EINTR)
            continue;
          warn ("Failed to read /proc/pid/cgroup");
          close (fd);
          return NULL;
        }
      break;
    }

  *out_length = ret;
  buffer[*out_length] = '\0';
  close (fd);

  if (*out_length > 5 &&  /* at least "0::/\n" */
      *out_length <= sizeof (buffer) - 1 && /* not too big */
      strncmp (buffer, "0::", 3) == 0 && /* must be a cgroupsv2 */
      buffer[*out_length - 1] == '\n') /* must end with a newline */
    return strdupx (buffer);

  warnx ("unexpected cgroups content, certificate matching only supports cgroup v2: '%s'", buffer);
  return NULL;
}

static
unsigned long long get_proc_pid_start_time (int dirfd)
{
  int fd = openat (dirfd, "stat", O_RDONLY);
  if (fd < 0)
    err (EX, "Failed to open /proc/pid/stat");

  char buffer[4096];
  ssize_t size = read (fd, buffer, sizeof buffer);
  if (size < 0)
    err (EX, "Failed to read /proc/pid/stat");
  close (fd);

  /* ensure this is NUL terminated */
  buffer[size] = '\0';

  /* start time is the token at index 19 after the '(process name)' entry - since only this
  * field can contain the ')' character, search backwards for this to avoid malicious
  * processes trying to fool us; See proc_pid_stat(5) */
  char *p = strrchr (buffer, ')');
  if (p == NULL)
    err (EX, "Failed to find process name in /proc/pid/stat");
  for (int i = 0; i <= 19; i++) /* NB: ')' is the first token */
    {
      p = strchr (p, ' ');
      if (p == NULL)
        err (EX, "Failed to find start time in /proc/pid/stat");
      ++p; /* skip over the space */
    }

  char *endptr;
  unsigned long long start_time = strtoull (p, &endptr, 10);
  if (*endptr != ' ')
    errx (EX, "Failed to parse start time in /proc/pid/stat from %s", p);
  return start_time;
}

/* valid_256_bit_hex_string:
 * @str: a string
 *
 * Ensures that str is a hexadecimal character string, exactly 64
 * characters in length.
 */
static bool
valid_256_bit_hex_string (const char *str)
{
  size_t length = strspn (str, "0123456789abcdef");

  return str[length] == '\0' && length == 64;
}

/**
 * read_cert_file:
 * @contents: a buffer to read the certificate into
 * @contents_size: the size of @contents
 *
 * Reads the contents of the certificate file into @contents (of size @contents_size).
 * The buffer must be large enough for the contents of the certificate file, plus
 * a nul terminator (which will be added).
 *
 * On success, the size of the certificate file (excluding nul
 * terminator) is returned.  This value is never 0.  On error, -1 is
 * returned with errno not guaranteed to be set (but a message will be
 * logged).
 */
static ssize_t
read_cert_file (const char *filename,
                char       *contents,
                size_t      contents_size)
{
  int dirfd = -1, filefd = -1;
  ssize_t result = -1;
  struct stat buf;
  ssize_t r;

  /* No tricky stuff, please */
  if (!valid_256_bit_hex_string (filename))
    {
      warnx ("tls-cert authentication token is invalid");
      goto out;
    }

  dirfd = open (CLIENT_CERTIFICATE_DIRECTORY, O_PATH | O_DIRECTORY | O_NOFOLLOW | O_CLOEXEC);
  if (dirfd == -1)
    {
      warn ("Failed to open " CLIENT_CERTIFICATE_DIRECTORY);
      goto out;
    }

  filefd = openat (dirfd, filename, O_RDONLY | O_NOFOLLOW);
  if (filefd == -1)
    {
      warn ("Failed to open certificate file %s/%s",
            CLIENT_CERTIFICATE_DIRECTORY, filename);
      goto out;
    }

  if (fstat (filefd, &buf) != 0)
    {
      warn ("Failed to stat certificate file %s/%s",
            CLIENT_CERTIFICATE_DIRECTORY, filename);
      goto out;
    }

  if (!S_ISREG (buf.st_mode))
    {
      warnx ("Could not read certificate: %s/%s is not a regular file",
             CLIENT_CERTIFICATE_DIRECTORY, filename);
      goto out;
    }

  if (buf.st_size == 0)
    {
      warnx ("Could not read certificate: %s/%s is empty",
             CLIENT_CERTIFICATE_DIRECTORY, filename);
      goto out;
    }

  /* Strictly less than, since we will add a nul */
  if (!(buf.st_size < contents_size))
    {
      warnx ("Insufficient space in read buffer for %s/%s",
             CLIENT_CERTIFICATE_DIRECTORY, filename);
      goto out;
    }

  do
    r = pread (filefd, contents, buf.st_size, 0);
  while (r == -1 && errno == EINTR);
  if (r == -1)
    {
      warn ("Could not read certificate file %s/%s",
            CLIENT_CERTIFICATE_DIRECTORY, filename);
      goto out;
    }
  if (r != buf.st_size)
    {
      warnx ("Read incomplete contents of certificate file %s/%s: %zu of %zu bytes",
             CLIENT_CERTIFICATE_DIRECTORY, filename, r, (size_t) buf.st_size);
      goto out;
    }

  contents[buf.st_size] = '\0';

  if (strlen (contents) != buf.st_size)
    {
      warnx ("Certificate file %s/%s contains nul characters",
             CLIENT_CERTIFICATE_DIRECTORY, filename);
      goto out;
    }

  result = buf.st_size;

out:
  if (filefd != -1)
    close (filefd);

  if (dirfd != -1)
    close (dirfd);

  return result;
}

static bool
sssd_map_certificate (const char *certificate, char** username)
{
  int result = false;
  sd_bus_error err = SD_BUS_ERROR_NULL;
  sd_bus *bus = NULL;
  sd_bus_message *user_obj_msg = NULL;
  const char *user_obj_path = NULL;
  int r;

  assert (username);
  assert (!*username);

  r = sd_bus_open_system (&bus);
  if (r < 0)
    {
      warnx ("Failed to connect to system bus: %s", strerror (-r));
      goto out;
    }

  /* sssd 2.6.1 introduces certificate validation against the configured CA. This version is in all supported distros */
  r = sd_bus_call_method (bus,
                          "org.freedesktop.sssd.infopipe",
                          "/org/freedesktop/sssd/infopipe/Users",
                          "org.freedesktop.sssd.infopipe.Users",
                          "FindByValidCertificate",
                          &err,
                          &user_obj_msg,
                          "s",
                          certificate);

  if (r < 0)
    {
      /* The error name is a bit confusing, and this is the common case; translate to readable error */
      if (sd_bus_error_has_name (&err, "sbus.Error.NotFound"))
        {
          warnx ("No matching user for certificate");
          goto out;
        }

      warnx ("Failed to map certificate to user: [%s] %s", err.name, err.message);
      goto out;
    }

  assert (user_obj_msg);

  r = sd_bus_message_read (user_obj_msg, "o", &user_obj_path);
  if (r < 0)
    {
      warnx ("Failed to parse response message: %s", strerror (-r));
      goto out;
    }

  debug ("certificate mapped to user object path %s", user_obj_path);

  r = sd_bus_get_property_string (bus,
                                  "org.freedesktop.sssd.infopipe",
                                  user_obj_path,
                                  "org.freedesktop.sssd.infopipe.Users.User",
                                  "name",
                                  &err,
                                  username);

  if (r < 0)
    {
      warnx ("Failed to map user object to name: [%s] %s", err.name, err.message);
      goto out;
    }

  assert (*username);
  debug ("mapped certificate to user %s", *username);
  result = true;

out:
  sd_bus_error_free (&err);
  sd_bus_message_unref (user_obj_msg);
  sd_bus_unref (bus);
  return result;
}

/**
 * cockpit_session_client_certificate_map_user
 *
 * Read the given certificate file, ensure that it belongs to our own cgroup, and ask
 * sssd to map it to a user. If everything matches as expected, return the user name.
 * Otherwise return %NULL, a warning message will already have been logged.
 */
char *
cockpit_session_client_certificate_map_user (const char *client_certificate_filename)
{
  char cert_pem[MAX_PEER_CERT_SIZE];
  char *sssd_user = NULL;

  /* read the certificate file from disk */
  if (read_cert_file (client_certificate_filename, cert_pem, sizeof cert_pem) < 0)
    {
      warnx ("No https instance certificate present");
      return NULL;
    }

  /* check if stdin is a Unix socket; then we are being called via cockpit-session@.service and need to
   * check the cgroup of our peer */
  struct ucred ucred;
  int ws_pid_dirfd;
  if (getsockopt (STDIN_FILENO, SOL_SOCKET, SO_PEERCRED, &ucred, &(socklen_t){sizeof ucred}) == 0)
    {
      debug ("unix socket mode, reading cgroup from peer pid %d", ucred.pid);
      ws_pid_dirfd = open_proc_pid (ucred.pid);
      int my_pid_dirfd = open_proc_pid (getpid ());
      unsigned long ws_start_time = get_proc_pid_start_time (ws_pid_dirfd);
      unsigned long my_start_time = get_proc_pid_start_time (my_pid_dirfd);
      debug ("peer start time: %lu, my start time: %lu", ws_start_time, my_start_time);
      close (my_pid_dirfd);

      /* Guard against pid recycling: If a malicious user captures ws, keeps the socket in a forked child and exits
       * the original pid, they can trap a different user to login, get the old pid (pointing ot their cgroup), and
       * capture their session. To prevent that, require that ws must have started earlier than ourselves. */
      if (my_start_time < ws_start_time)
        {
          warnx ("start time of this process (%lu) is older than cockpit-ws (%lu), pid recycling attack?",
                  my_start_time, ws_start_time);
          close (ws_pid_dirfd);
          return NULL;
        }
    }
  else {
    ws_pid_dirfd = open_proc_pid (getpid ());
    debug ("failed to read stdin peer credentials: %m; not in socket mode, reading cgroup from my own pid %d", ws_pid_dirfd);
  }

  size_t ws_cgroup_length;
  char *ws_cgroup = read_proc_pid_cgroup (ws_pid_dirfd, &ws_cgroup_length);
  close (ws_pid_dirfd);
  if (ws_cgroup == NULL)
    {
      warnx ("Could not determine cgroup of this process");
      return NULL;
    }
  /* A simple prefix comparison is appropriate here because ws_cgroup
   * will contain exactly one newline (at the end), and the expected
   * value of ws_cgroup is on the first line in cert_pem.
   */
  if (strncmp (cert_pem, ws_cgroup, ws_cgroup_length) != 0)
    {
      warnx ("This client certificate is only meant to be used from another cgroup");
      free (ws_cgroup);
      return NULL;
    }
  free (ws_cgroup);

  /* ask sssd to map cert to a user */
  if (!sssd_map_certificate (cert_pem + ws_cgroup_length, &sssd_user))
    return NULL;

  return sssd_user;
}
